#!/usr/bin/perl -I/home/szpari/csahserver/FFT/lib/perl5/site_perl/5.12.1/x86_64-linux-thread-multi/
#!/usr/bin/perl -I/home/szpari/programs/Perl
#-I/home/apu/ZOLI/FT_CHARGE/lib/perl/5.8.7/
#-I/home/szpari/programs/Perl

print STDERR "
#######################################################################
# ft_charge_evd.pl [-n<max_correl_dist>] [-m 'correl'|'cft'|'fft']    #
#                  [-w<window_length>] [-t<step>] [-e<evd_parm_file>] #
#                  [-p<max_P_value] [-a<minamp>] [-f<lfreq:hfreq>]    #
#                  [-o<html_output_file>] [-s]                        #
#                  < infile.fst > outfile                             #
#                                                                     #
# Identifies repetitive charge patterns in the input sequence(s).     #
# The input file may contain multiple fasta entries.  Options:        #
# -n defines the maximum sequential distance for charge correlation   #
#    search (adjusted to next power of 2 for FFT). Default: 'auto'    #
#    (length*0.75 for each sequence)                                  #
# -m mode of operation: 'correl' returns the charge correlation       #
#    function while 'cft' the spectrum of it and 'fft' that of the    #
#    sequence itself (default: cft)                                   #
# -w defines the window length or range for segmental search          #
#    (default: 16-64)                                                 #
# -t controls the sliding step between the windows (default: 1)       #
# -p maximum P value to report (default: 0.01)                        #
# -a minimum amplitude to report (default: 10)                        #
# -f frequency range to report (default: 1/6:1/9)                     #
# -o HTML output file (table with CSAH parameters)                    #
# -s if specified, will write full otuput including sequences         #
#                                                                     #
# Note: uses FFT.pm obtainable from search.cpan.org/dist/Math-FFT/    #
#                                                                     #
# Version 4.0                     06.04.2011.                         #
#######################################################################

";
$|=1;

use Getopt::Std;
$opt_n='auto';
$opt_m="CFT";
$opt_o="";
$opt_p=0.05;
$opt_a=7;
$opt_w="32-64";
$opt_t=1;
$opt_f="1/6:1/9";
$opt_e="/home/szpari/bin/ftcharge_evdtable.txt";
getopts('m:n:o:p:a:w:t:e:f:shv') || die "Wrong options!\n";
exit if $opt_h;
if ($opt_v){
    $V=1;
    $vout="high";
}
else{$vout="low"}


use Math::FFT;
$MINSLEN=8;


($LFREQ,$HFREQ)=split(/:/,$opt_f);
$LFREQ=eval("$LFREQ");
$HFREQ=eval("$HFREQ");
if ($HFREQ < $LFREQ){$tr=$LFREQ; $LFREQ=$HFREQ; $HFREQ=$tr}

if ($opt_o ne ""){
 open (HO,">$opt_o") || die "Cannot create html output file $opt_o!\n";
 print HO "<HTML>
<HEAD><TITLE>filtered ft_charge.pl output</TITLE></HEAD>
<BODY>

<FONT FACE=\"Courier New, Courier\">
<TABLE CELLSPACING=0 BORDER=1><TR BGCOLOR=\"lightgreen\"><TH>Sequence<br>ID</TH><TH>CSAH</TH><TH>P-value</TH><TH>Maximum<br>amplitude</TH><TH>Frequency</TH><TH>Sequence<TH></TR>
";

}

if (($opt_m ne "c") && ("correl" =~ /$opt_m/i)){$MODE="CORREL"}
elsif ("fft"=~/$opt_m/i){$MODE="FFT"}
else{$MODE="CFT"}

$MINAMP=$opt_a;
$MA=$MINAMP;$MA=~s/[A-Z]//gi;

print "
Options set:
============
max_correl_dist      $opt_n
mode                 $MODE
window size          $opt_w
sliding step size    $opt_t
EVD parameter file   $opt_e
maximum P-value      $opt_p
minimum amplitude    $opt_a
verbosity level      $vout

sequence_segment_id           P-value  max_ampl  max_pos";
print "  sequence" if $opt_f;
print "\n";

# Reading EVD parameters
open (EP,"$opt_e") || die ("Cannot open EVD paremeter file: $opt_e\n");
while(<EP>){
    chomp;
    $_=~s/^ +//;
    ($len,$P,$N,$EMODE,$loc,$scale,$shape)=split(/ +/,$_);
    $stype=$len."_".$P."_".$N;
    $EVD_LOC{$stype}{$EMODE}=$loc;
    $EVD_SCALE{$stype}{$EMODE}=$scale;
    $EVD_SHAPE{$stype}{$EMODE}=$shape;
}
close(EP);




print STDERR "MINAMP: $MINAMP\n" if $V;

%charge=('E',-1,'D',-1,
         'H',0.5,
         'K',1,'R',1);

$seq="";
$sn=0;

$maxN=0;
$maxM=0;
$allsegments=0;

$hits=0;
# Analyzing seqs 'on-the-fly'
while(<>){
 chop;
 $_=~s/\r//g;
 if ($_=~/^>/){
     if ($sn != 0){&analyze_seq}
     $sn++;
     #$seqid[$sn]=substr($_,1,10);
     $seq="";
     #$seqid=$_;
     ($seqid,$def) = split(' ',$_,2);
     $seqid =~ s/^>//;
     #$seqid=~s/\>//g;
     #$seqid=~s/\|/\_/g;
     #$seqid=~s/ .*$//;
}
 else{
     $seq.=$_;
 }
}
if ($seq ne ""){&analyze_seq}


print STDERR "\nRead and processed $sn sequences, found $hits hits\n";

# closing HTML output
if ($opt_o ne ""){
print HO "</TABLE>

</FONT>
</BODY></HTML>
";
close(HO);
}

# End program
exit;

sub analyze_seq{

#printf STDERR ("Processing sequence No.%7d\r",$sn); 
#print STDERR "sequence: $seq\n";

    $STEP=$opt_t;
    $WINDOW=$opt_w;
    if ($opt_w =~ /full/i){
	$WIN_FROM=length($seq);
	$WIN_TO=length($seq);
	$STEP=length($seq);
    }
    elsif ($opt_w =~ /\-/){
        ($WIN_FROM,$WIN_TO)=split(/\-/,$opt_w);
    }
    else{
        $WIN_FROM=$opt_w;
        $WIN_TO=$opt_w;
    }


    while ($WIN_FROM <= $WIN_TO){
    $WINDOW=$WIN_FROM;
    print STDERR "Actual window: $WINDOW\n" if $opt_v;
    #print "Actual window: $WINDOW  \n";

    for ($sp=0; $sp < length($seq); $sp+=$STEP){
    ##for ($sp=0; $sp < 1; $sp++){
	@OUT_FN=();
        @U=();
	$round=0;
	##$segment=$seq;
	$segment=substr($seq,$sp,$WINDOW);
	$start=$sp+1;$end=$sp+$WINDOW;
	if ($end > length($seq)){$end=length($seq)}
	if (length($segment) > $MINSLEN){
	$segmentid=$seqid."_$start-$end";

	print STDERR "$segmentid : >\n$segment\n" if $V;

	$SEGID[$allsegments]=$segmentid if $WRITEPLOT;
	$allsegments++;
	&do_ft_charge;
	$stype=getstype($segment);
	&calc_P_value($stype);
#	print STDERR "segmentid $segmentid segment $segment \n>>stype $stype Pval $P_VALUE MAXVAL $MAX_VALUE[0] MPV $MPV\n";
	$MP=simplify($MAX_POS[0]);
        if ($MP !~ /\/0$/){$MPV=1/eval($MP)} else {$MPV=0};
#	print STDERR "segmentid $segmentid segment $segment \n>>stype $stype Pval $P_VALUE MAXVAL $MAX_VALUE[0] MPV $MPV round $round\n";
	if (($P_VALUE <= $opt_p) && ($MAX_VALUE[0] >= $opt_a) && ($MPV > $LFREQ) && ($MPV < $HFREQ)){
	    print ("$segmentid");
	    for ($spaces=30; $spaces > length($segmentid); $spaces--){print " "}
            $MPO=$MP;$MPO=~s/(^[0-9]+)\/([0-9]+)$/$2\/$1/g;
	    printf ("%7.3f   %7.3f   %6s",$P_VALUE,$MAX_VALUE[0],$MPO);
	    print "  $segment" if $opt_f;
	    print "\n";
	    $hits++;
            if ($opt_o ne ""){
                $oid=$seqid;$oid=~s/>//;
                $os=$segment;
                $os=~s/([ED])/<font color=red>$1<\/font>/g;
                $os=~s/([RK])/<font color=blue>$1<\/font>/g;
                printf HO ("<tr><td align=center>$oid<td align=center><nobr>$start-$end</nobr><td align=center>%7.4f<td align=center>%7.2f<td align=center>$MPO<td align=center>$os</tr>\n",$P_VALUE,$MAX_VALUE[0]);

              }
	}
    }#_if length(segment) > MIN
    }
    $WIN_FROM*=2;
    }#while WIN_FROM < WIN_TO

}# sub analyze_seq



sub do_ft_charge{
    $maxN=0;$maxM=0;

# charge correlation function
if ($MODE=~/C/){
    print STDERR "Calculating the charge correlation function... (for $segmentid) " if $V;

    $seqlen=length($segment);
    if ($opt_n eq 'auto'){$N=$seqlen*0.75}
    else {$N=$opt_n}

    if ($N > $maxN){$maxN=$N}
    print STDERR "maxN: $maxN ($opt_n)\n" if $V;
    for ($n=1; $n<=$N; $n++){
	$U[$n]=0;
	for ($i=0; $i < length($segment); $i++){
	    $U[$n]+=$charge{substr($segment,$i,1)}*$charge{substr($segment,$i+$n,1)};
	}
    }

print STDERR "Done\n" if $V;
}#_if $MODE CORREL or CFT

# Doing FFT now
# Adjusting to next power of 2
$m=1;
if ($MODE eq "FFT"){
    $maxlen=length($segment);
    while ($m < $maxlen){$m*=2}
}
else{
    while ($m < $maxN){$m*=2}
}
if ($m > $maxM){$maxM=$m}

if ($MODE =~ /FT/){
    print STDERR "Doing FFT... (for $segmentid)" if $V;


    # FFT using the FFT module
    @toft=();
    my $data="";

    for ($x=0; $x < $m; $x++){
	#$toft[$x]=$U[$x+1] if $MODE eq "CFT";
        if ($MODE eq "CFT"){
         if ($U[$x+1] =~ /[0-9]/){$toft[$x]=$U[$x+1]}
         else{$toft[$x]=0};
        }
	elsif ($MODE eq "FFT"){
	    $toft[$x]=$charge{substr($segment,$x,1)};
	}
    }
    $data=\@toft;
    my $fft = new Math::FFT($data);
    $coeff=$fft->rdft();
    $spectrum=$fft->spctrm();


}

# Storing what needs to be stored if OUT_FN
# & calculating value and position of maximums 

$MAX_VALUE[$round]=0;
$MAX_POS[$round]=-1;
#print STDERR "GETTING MAX for round $round ($segmentid): ";
if ($MODE eq "CORREL"){

    for ($x=1; $x <= $N; $x++){
	$OUT_FN[$round][$x]=$U[$x];
	if (abs($U[$x]) > $MAX_VALUE[$round]){
	    $MAX_VALUE[$round]=$U[$x];
	    $MAX_POS[$round]=$x;
	}

	if ($WRITEPLOT){
	    $OUT_FN{$segmentid}[$x]=$U[$x];
	}
    }
}
# this is the same for FT modes
else{
    for ($x=0; $x < $m; $x++){
	$OUT_FN[$round][$x]=$spectrum->[$x];
	if ($spectrum->[$x] > $MAX_VALUE[$round]){
	    $MAX_VALUE[$round]=$spectrum->[$x];
	    #$MAX_POS[$round]="1/".$x;
	    $MAX_POS[$round]=$m."/".$x;
	    $MAX_COEFF[$round]=$coeff->[$x];
	}

	if ($WRITEPLOT){
	    $OUT_FN{$segmentid}[$x]=$spectrum->[$x];
	}
    }
}



print STDERR "Done\n" if $V;


}#_sub do_ft_charge

#######################################################

sub getstype{
    my $sgmnt=shift;

    $len=length($sgmnt);
    # getting length
    $minld=$len;
    foreach $sl (16,32,64,128){
	$ld=abs($len-$sl);
	if ($ld <= $minld){
	    $minld=$ld;
	    $slen=$sl;
	}
    }
    # getting charge percentage
    my $aa;
    my $sP=0; my $sN=0;
    foreach $aa (split(//,$sgmnt)){
	if ($charge{$aa} > 0){$sP+=$charge{$aa}};
	if ($charge{$aa} < 0){$sN-=$charge{$aa}};
    }
    $sP=0.5+(10*$sP/$len);
    $sN=0.5+(10*$sN/$len);
    $sP=~s/\..*$//;$sP*=10;
    $sN=~s/\..*$//;$sN*=10;

    if ($sP < 10){$sP=10};if ($sP > 90){$sP=90};
    if ($sN < 10){$sN=10};if ($sN > 90){$sN=90};

    my $stype=$slen."_".$sP."_".$sN;
    return $stype;

}#_sub getstype

sub calc_P_value{

    my $stype=shift;
    my $loc=$EVD_LOC{$stype}{$MODE};
    my $scale=$EVD_SCALE{$stype}{$MODE};
    my $shape=$EVD_SHAPE{$stype}{$MODE};
    my $x=$MAX_VALUE[0];

    ##print STDERR "stype $stype loc $loc scale $scale shape $shape\n";
    if ($scale == 0){$scale=10e-10}
    if ($shape == 0){$shape=10e-10}
    my $z = ($x - $loc) / $scale;
    $P_VALUE = 1 - exp(-(1+$shape*$z)**(-1/$shape));

}#_sub calc_P_value

#######################################################
sub FFT_anal{

    if ($opt_s ne ""){
	%amps=();
	$avg=0;
	$stdev=0;
	for ($x=0; $x < $m; $x++){
	    #$spectrum->[$x]/=$seqlen[$s];
	    $avg+=$spectrum->[$x];
	}
	$avg/=$m;
	for ($x=0; $x < $m; $x++){
	    $stdev+=($spectrum->[$x]-$avg)*($spectrum->[$x]-$avg);
	}
	$stdev=sqrt($stdev/$m);

	print STDERR "$seqid[$s] > AVG  $avg STDEV $stdev\n" if $V;
    }

    for ($x=0; $x < $m; $x++){
	$ft[$x]=$spectrum->[$x];

        # Identifying peaks: a simple value-based method
	if ($opt_s ne ""){
	    if ($MINAMP =~ /a/i){
		if ($ft[$s][$x] >= $avg*$MA){
		    $amps{$x}=$ft[$s][$x];
		}
	    }
	    elsif ($MINAMP =~ /d/i){
		if ($ft[$s][$x] >= $stdev*$MA){
		    print STDERR "$ft[$s][$x] >= $stdev * $MINAMP\n" if $V;
		    $amps{$x}=$ft[$s][$x];
		}
	    }
	    else{
		if ($ft[$s][$x] >= $MA){
		    $amps{$x}=$ft[$s][$x];
		}
	    }

	}
    }

    if ($opt_s ne ""){
	print SF ($seqid[$s]);
	print SF " "x(40-length($seqid[$s]));
	printf SF (" %5d ",$seqlen[$s]);
        foreach $peak (sort {$amps{$b} <=> $amps{$a}} keys %amps){ 
	    printf SF ("| %7.3f %7.1f ",$peak/$m,$amps{$peak}) if $MODE eq "CFT";
	    printf SF ("| %7.3f %7.1f ",$peak/$m,$amps{$peak}*100) if $MODE eq "FFT";

	}
	print SF "\n";
    }

}#_sub FFT_anal

#######################################################
# Simplifies a fraction given as $a."/".$b
sub simplify{

    $frac=shift;
    ($a,$b)=split(/\//,$frac);
    $simp=$frac;
    for ($D=$b; $D>0; $D--){
	$A=$a/$D;
	$Ai=$A;
	$Ai=~s/\..*$//;
	$B=$b/$D;
	$Bi=$B;
	$Bi=~s/\..*$//;

	if (($A eq $Ai) && ($B eq $Bi)){
	    $simp=$A."/".$B;
	    last;
	}
    }

    return $simp;

}#_sub simplify


